from .InfoTab import InfoTab
from .SplitsTab import SplitsTab
from .MLTab import MLTab
from .ResultsTab import ResultsTab
from .AggregatedResultsTab import AggregatedResultsTab
from .InterpretTab import InterpretTab
